## protocol.py
# IRC low-level protocol constants and helpers.
import re
import collections

## Own definitions.

# While this *technically* is supposed to be 143/994, I've yet to see a server that actually uses those.
DEFAULT_PORT = 6667
DEFAULT_TLS_PORT = 6697

FALLBACK_ENCODING = 'iso-8859-1'

## Limits.
CHANNEL_LIMITS_GROUPS = { '#': frozenset('#&'), '&': frozenset('#&') }
CHANNEL_LIMITS = { frozenset('#&'): 10 }
LIST_LIMIT_GROUPS = { 'b': frozenset('b') }
LIST_LIMITS = { frozenset('b'): None }
PARAMETER_LIMIT = 15
MESSAGE_LENGTH_LIMIT = 512
CHANNEL_LENGTH_LIMIT = 200
NICKNAME_LENGTH_LIMIT = 8
TOPIC_LENGTH_LIMIT = 450

## Defaults.

BEHAVIOUR_NO_PARAMETER = 'noparam'
BEHAVIOUR_PARAMETER = 'param'
BEHAVIOUR_PARAMETER_ON_SET = 'param_set'
BEHAVIOUR_LIST = 'list'

CHANNEL_MODES = { 'o', 'p', 's', 'i', 't', 'n', 'b', 'v', 'm', 'r', 'k', 'l' }
CHANNEL_MODES_BEHAVIOUR = { BEHAVIOUR_LIST: { 'b' }, BEHAVIOUR_PARAMETER: { 'o', 'v' }, BEHAVIOUR_PARAMETER_ON_SET: { 'k', 'l' }, BEHAVIOUR_NO_PARAMETER: { 'p', 's', 'i', 't', 'n', 'm', 'r' } }
CHANNEL_PREFIXES = { '#', '&' }
CASE_MAPPINGS = { 'ascii', 'rfc1459', 'strict-rfc1459' }
CASE_MAPPING = 'rfc1459'
NICKNAME_PREFIXES = collections.OrderedDict([('@', 'o'), ('+', 'v')])
USER_MODES = { 'i', 'w', 's', 'o' }
# Maybe one day, user modes will have parameters...
USER_MODES_BEHAVIOUR = { BEHAVIOUR_NO_PARAMETER: { 'i', 'w', 's', 'o' } }

## Message parsing.

FORBIDDEN_CHARACTERS = { '\r', '\n', '\0' }

LINE_SEPARATOR = '\r\n'
MINIMAL_LINE_SEPARATOR = '\n'
USER_SEPARATOR = '!'
HOST_SEPARATOR = '@'

PRIVATE_CHANNEL_SIGIL = '@'
SECRET_CHANNEL_SIGIL = '*'
PUBLIC_CHANNEL_SIGIL = '='

ARGUMENT_SEPARATOR = re.compile(' +', re.UNICODE)
COMMAND_PATTERN = re.compile('^([a-zA-Z]+|[0-9]+)$', re.UNICODE)
TRAILING_PREFIX = ':'


## Errors.

class ProtocolViolation(Exception):
    """ An error that occurred while parsing or constructing an IRC message that violates the IRC protocol. """
    def __init__(self, msg, message):
        super().__init__(msg)
        self.irc_message = message


## Message parsing and construction.

# Construction.

def construct(command, *params, source=None):
    """ Construct a raw IRC message. """
    # Sanity check for command.
    command = str(command)
    if not COMMAND_PATTERN.match(command):
        raise ProtocolViolation('The constructed command does not follow the command pattern ({pat})'.format(pat=COMMAND_PATTERN.pattern), message=command)
    message = command.upper()

    # Add parameters.
    if not params:
        message += ' '
    for idx, param in enumerate(params):
        # Trailing parameter?
        if ' ' in param:
            if idx + 1 < len(params):
                raise ProtocolViolation('Only the final parameter of an IRC message can be trailing and thus contain spaces.', message=param)
            message += ' ' + TRAILING_PREFIX + param
        # Regular paramter.
        else:
            message += ' ' + param

    # Prepend source.
    if source:
        message = ':' + source + ' ' + message

    # Sanity check for characters.
    if any(ch in message for ch in FORBIDDEN_CHARACTERS):
        raise ProtocolViolation('The constructed message contains forbidden characters ({chs}).'.format(chs=', '.join(FORBIDDEN_CHARACTERS)), message=message)

    # Sanity check for length.
    message += LINE_SEPARATOR
    if len(message) > MESSAGE_LENGTH_LIMIT:
        raise ProtocolViolation('The constructed message is too long. ({len} > {maxlen})'.format(len=len(message), maxlen=MESSAGE_LENGTH_LIMIT), message=message)

    return message


# Validation.

def normalize(input, case_mapping='rfc1459'):
    """ Normalize input according to case mapping. """
    if case_mapping not in CASE_MAPPINGS:
        raise ProtocolViolation('Unknown case mapping ({})'.format(case_mapping))

    input = input.lower()

    if case_mapping in ('rfc1459', 'rfc1459-strict'):
        input = input.replace('{', '[').replace('}', ']').replace('|', '\\')
    if case_mapping == 'rfc1459':
        input = input.replace('~', '^')

    return input

def equals(left, right, case_mapping='rfc1459'):
    """ Determine whether or not the given nicknames are equal in the given case mapping. """
    if case_mapping not in CASE_MAPPINGS:
        raise ProtocolViolation('Unknown case mapping ({})'.format(case_mapping))

    return normalize(left, case_mapping) == normalize(right, case_mapping)


# Parsing.

def parse(line, encoding='utf-8'):
    """
    Parse given line into IRC message structure.
    Returns a tuple of (source, command, parameters), where parameters is a list.
    """
    # Decode message.
    try:
        message = line.decode(encoding)
    except UnicodeDecodeError:
        # Try our fallback encoding.
        message = line.decode(FALLBACK_ENCODING)

    # Sanity check for message length.
    if len(message) > MESSAGE_LENGTH_LIMIT:
        raise ProtocolViolation('The received message is too long. ({len} > {maxlen})'.format(len=len(message), maxlen=MESSAGE_LENGTH_LIMIT), message=message)

    # Strip message separator.
    if message.endswith(LINE_SEPARATOR):
        message = message[:-len(LINE_SEPARATOR)]
    elif message.endswith(MINIMAL_LINE_SEPARATOR):
        message = message[:-len(MINIMAL_LINE_SEPARATOR)]

    # Sanity check for forbidden characters.
    if any(ch in message for ch in FORBIDDEN_CHARACTERS):
        raise ProtocolViolation('The received message contains forbidden characters ({chs}).'.format(chs=', '.join(repr(x) for x in FORBIDDEN_CHARACTERS)), message=message)

    # Extract message sections.
    # Format: (:source)? command parameter*
    try:
        if message.startswith(':'):
            source, command, raw_params = ARGUMENT_SEPARATOR.split(message, 2)
            source = source[1:]
        else:
            command, raw_params = ARGUMENT_SEPARATOR.split(message, 1)
            source = None
    except ValueError:
        raise ProtocolViolation('Improper IRC message format: not enough elements.')

    # Sanity check for command.
    if not COMMAND_PATTERN.match(command):
        raise ProtocolViolation('The received command ({message}) is not a valid IRC command.'.format(message=command))

    # Extract parameters properly.
    # Format: (word|:sentence)*

    # Only parameter is a 'trailing' sentence.
    if raw_params.startswith(TRAILING_PREFIX):
        params = [ raw_params[len(TRAILING_PREFIX):] ]
    # We have a sentence in our parameters.
    elif ' ' + TRAILING_PREFIX in raw_params:
        index = raw_params.find(' ' + TRAILING_PREFIX)

         # Get all single-word parameters.
        params = ARGUMENT_SEPARATOR.split(raw_params[:index].rstrip(' '))
        # Extract last parameter as sentence
        params.append(raw_params[index + len(TRAILING_PREFIX) + 1:])
    # We have some parameters, but no sentences.
    elif raw_params:
        params = ARGUMENT_SEPARATOR.split(raw_params)
    # No parameters.
    else:
        params = []

    # Commands can be either [a-zA-Z]+ or [0-9]+.
    # In the former case, force it to yppwecase.
    # In the latter case (a numeric command), try to represent it as such.
    try:
        command = int(command)
    except ValueError:
        command = command.upper()

    # Return parsed message.
    return (source, command, params)

def parse_user(raw):
    """ Parse nick(!user(@host)?)? structure. """
    nick = raw
    user = None
    host = None

    # Attempt to extract host.
    if HOST_SEPARATOR in raw:
        raw, host = raw.split(HOST_SEPARATOR)
    # Attempt to extract user.
    if USER_SEPARATOR in raw:
        nick, user = raw.split(USER_SEPARATOR)

    return nick, user, host

def parse_user_modes(modes, current):
    """ Parse user mode changes and return updated mode dictionary. """
    return parse_modes(modes, current, behaviour=USER_MODES_BEHAVIOUR)

def parse_channel_modes(modes, current, behaviour=CHANNEL_MODES_BEHAVIOUR):
    """ Parse channel mode changes and return updated mode dictionary. """
    return parse_modes(modes, current, behaviour=behaviour)

def parse_modes(modes, current, behaviour):
    """ Parse mode change string(s) and return updated dictionary. """
    current = current.copy()

    # Iterate in a somewhat odd way over the list because we want to modify it during iteration.
    i = 0
    while i < len(modes):
        piece = modes[i]
        add = True
        sigiled = False

        for mode in piece:
            # Set mode to addition or deletion of modes.
            if mode == '+':
                add = True
                sigiled = True
                continue
            if mode == '-':
                add = False
                sigiled = True
                continue

            # Find mode behaviour.
            for type, affected in behaviour.items():
                if mode in affected:
                    break
            else:
                # If we don't have a behaviour for this mode, assume it has no parameters...
                type = BEHAVIOUR_NO_PARAMETER

            # Don't parse modes that are meant for list retrieval.
            if type == BEHAVIOUR_LIST and not sigiled:
                continue

            # Do we require a parameter?
            if type in (BEHAVIOUR_PARAMETER, BEHAVIOUR_LIST) or (type == BEHAVIOUR_PARAMETER_ON_SET and add):
                # Do we _have_ a parameter?
                if i + 1 == len(modes):
                    raise ProtocolViolation(
                        'Attempted to parse mode with parameter ({s}{mode}) but no parameters left in mode list.'.format(
                        mode=mode, s='+' if add else '-'), ' '.join(modes))
                param = modes.pop(i + 1)

            # Now update the actual mode dict with our new values.
            if type in (BEHAVIOUR_PARAMETER, BEHAVIOUR_LIST):
                # Add/remove parameter from list.
                if add:
                    if mode not in current:
                        current[mode] = []
                    current[mode].append(param)
                else:
                    if mode in current and param in current[mode]:
                        current[mode].remove(param)
            elif type == BEHAVIOUR_PARAMETER_ON_SET and add:
                # Simply set parameter.
                current[mode] = param
            else:
                # Simply add/remove option.
                if add:
                    current[mode] = True
                else:
                    if mode in current:
                        del current[mode]

        # Onto the next mode.
        i += 1

    return current




## ctcp.py
# Client-to-Client-Protocol (CTCP) support.
from .. import client
from .. import protocol

__all__ = [ 'CTCPSupport' ]


CTCP_DELIMITER = '\x01'
CTCP_ESCAPE_CHAR = '\x16'


class CTCPSupport(client.BasicClient):
    """ Support for CTCP messages. """

    ## Callbacks.

    def on_ctcp(self, by, target, what):
        pass

    def on_ctcp_reply(self, by, target, what, response):
        pass

    def on_ctcp_version(self, by, target):
        """ Built-in CTCP version as some networks seem to require it. """
        import pydle

        version = '{name} v{ver}'.format(name=pydle.__name__, ver=pydle.__version__)
        self.ctcp_reply(by, 'VERSION', version)


    ## IRC API.

    def ctcp(self, target, query):
        """ CTCP request from target. """
        if self.is_channel(target) and not self.in_channel(target):
            raise client.NotInChannel('Not in channel {}'.format(target))

        self.message(target, construct_ctcp(query))

    def ctcp_reply(self, target, query, response):
        """ CTCP reply to target. """
        if self.is_channel(target) and not self.in_channel(target):
            raise client.NotInChannel('Not in channel {}'.format(target))

        self.notice(target, construct_ctcp(query, response))


    ## Handler overrides.

    def on_raw_privmsg(self, source, params):
        """ Modify PRIVMSG to redirect CTCP messages. """
        nick, user, host = protocol.parse_user(source)
        target, message = params

        if is_ctcp(message):
            if nick in self.users:
                self._sync_user(nick, user, host)
            type = parse_ctcp_query(message)

            # Find dedicated handler if it exists.
            attr = 'on_ctcp_' + type.lower()
            if hasattr(self, attr):
                getattr(self, attr)(nick, target)
            else:
                # Invoke global handler.
                self.on_ctcp(nick, target, type)
        else:
            super().on_raw_privmsg(source, params)

    def on_raw_notice(self, source, params):
        """ Modify NOTICE to redirect CTCP messages. """
        nick, user, host = protocol.parse_user(source)
        target, message = params

        if is_ctcp(message):
            if nick in self.users:
                self._sync_user(nick, user, host)
            type, response = parse_ctcp_response(message)

            # Find dedicated handler if it exists.
            attr = 'on_ctcp_' + type.lower() + '_reply'
            if hasattr(self, attr):
                getattr(self, attr)(user, target, response)
            else:
                # Invoke global handler.
                self.on_ctcp_reply(user, target, type, response)
        else:
            super().on_raw_notice(source, params)


## Helpers.

def is_ctcp(message):
    """ Check if message follows the CTCP format. """
    return message.startswith(CTCP_DELIMITER) and message.endswith(CTCP_DELIMITER)

def construct_ctcp(*parts):
    """ Construct CTCP message. """
    message = ' '.join(parts)
    message = message.replace('\0', CTCP_ESCAPE_CHAR + '0')
    message = message.replace('\n', CTCP_ESCAPE_CHAR + 'n')
    message = message.replace('\r', CTCP_ESCAPE_CHAR + 'r')
    message = message.replace(CTCP_ESCAPE_CHAR, CTCP_ESCAPE_CHAR + CTCP_ESCAPE_CHAR)
    return CTCP_DELIMITER + message + CTCP_DELIMITER

def parse_ctcp_query(query):
    """ Strip and de-quote CTCP messages. """
    query = query.strip(CTCP_DELIMITER)
    query = query.replace(CTCP_ESCAPE_CHAR + '0', '\0')
    query = query.replace(CTCP_ESCAPE_CHAR + 'n', '\n')
    query = query.replace(CTCP_ESCAPE_CHAR + 'r', '\r')
    query = query.replace(CTCP_ESCAPE_CHAR + CTCP_ESCAPE_CHAR, CTCP_ESCAPE_CHAR)
    return query

def parse_ctcp_response(response):
    """ Strip and de-quote CTCP response. """
    response = parse_ctcp_query(response)
    return response.split(' ', 1)



import sys
import os.path as path
import itertools

import socket
import ssl
import select

import errno
import collections
import threading

from . import protocol

__all__ = [ 'BUFFER_SIZE', 'TIMEOUT', 'NotConnected', 'NoMessageAvailable', 'Connection' ]


DEFAULT_CA_PATHS = {
    'linux': '/etc/ssl/certs',
    'linux2': '/etc/ssl/certs',
    'freebsd': '/etc/ssl/certs'
}

BUFFER_SIZE = 4096
TIMEOUT = 0.5


class NotConnected(Exception):
    pass

class NoMessageAvailable(Exception):
    pass


class Connection:
    """ A TCP connection over the IRC protocol. """
    CONNECT_TIMEOUT = 10

    def __init__(self, hostname, port, tls=False, tls_verify=True, encoding='utf-8', tls_certificate_file=None, tls_certificate_keyfile=None, ping_timeout=240):
        self.hostname = hostname
        self.port = port
        self.ping_timeout = ping_timeout
        self.encoding = encoding

        self.tls = tls
        self.tls_context = None
        self.tls_verify = tls_verify
        self.tls_certificate_file = tls_certificate_file
        self.tls_certificate_keyfile = tls_certificate_keyfile

        self.timer = None
        self.timer_lock = threading.RLock()
        self.socket = None
        self.socket_lock = threading.RLock()
        self.buffer = None
        self.buffer_lock = threading.RLock()
        self.message_queue = None
        self.message_lock = threading.RLock()

    def connect(self):
        """ Connect to target. """
        with self.socket_lock:
            # Create regular socket.
            self.socket = socket.create_connection((self.hostname, self.port), timeout=self.CONNECT_TIMEOUT)

            # Wrap it in a TLS socket if we have to.
            if self.tls:
                self.setup_tls()

            self.socket.settimeout(TIMEOUT)
            # Enable keep-alive.
            if hasattr(socket, 'SO_KEEPALIVE'):
                self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # Have the socket return an error instead of signaling SIGPIPE.
            if hasattr(socket, 'SO_NOSIGPIPE'):
                self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_NOSIGPIPE, 1)

        # Reset message buffer and queue.
        with self.buffer_lock:
            self.buffer = b''
        with self.message_lock:
            self.message_queue = collections.deque()

    def setup_tls(self):
        """ Transform our regular socket into a TLS socket. """
        # Set up context.
        self.tls_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

        # Load client/server certificate.
        if self.tls_certificate_file:
            self.tls_context.load_cert_chain(self.tls_certificate_file, self.tls_certificate_keyfile)

        # Set some relevant options.
        # No server should use SSLv2 any more, it's outdated and full of security holes.
        # Disable compression to counter the CRIME attack. (https://en.wikipedia.org/wiki/CRIME_%28security_exploit%29)
        for opt in [ 'NO_SSLv2', 'NO_COMPRESSION']:
            if hasattr(ssl, 'OP_' + opt):
                self.tls_context.options |= getattr(ssl, 'OP_' + opt)

        # Set TLS verification options.
        if self.tls_verify:
            # Set our custom verification callback.
            if hasattr(self.tls_context, 'set_servername_callback'):
                self.tls_context.set_servername_callback(self.verify_tls)

            # Always set default root certificate storage paths.
            self.tls_context.set_default_verify_paths()

            # Try OS-specific paths too in case the above failed.
            if sys.platform in DEFAULT_CA_PATHS and path.isdir(DEFAULT_CA_PATHS[sys.platform]):
                self.tls_context.load_verify_locations(ca_path=DEFAULT_CA_PATHS[sys.platform])

            # Enable verification.
            self.tls_context.verify_mode = ssl.CERT_REQUIRED
            # Verify the entire chain if possible.
            if hasattr(self.tls_context, 'verify_flags'):
                self.tls_context.verify_flags = ssl.VERIFY_CRL_CHECK_CHAIN

        # Wrap socket into a snuggly TLS blanket.
        self.socket = self.tls_context.wrap_socket(self.socket,
            # Send hostname over SNI, but only if our TLS library supports it.
            server_hostname=self.hostname if ssl.HAS_SNI else None)

        # And verify the peer here if our TLS library doesn't have callback functionality.
        if self.tls_verify and not hasattr(self.tls_context, 'set_servername_callback'):
            self.verify_tls(self.socket, self.hostname, self.tls_context, as_callback=False)

    def verify_tls(self, socket, hostname, context, as_callback=True):
        """
        Verify a TLS connection. Return behaviour is dependent on the as_callback parameter:
            - If True, a return value of None means verification succeeded, else it failed.
            - If False, a return value of True means verification succeeded, an exception or False means it failed.
        """
        cert = socket.getpeercert()

        try:
            ssl.match_hostname(cert, hostname)
        except ssl.CertificateError:
            if not as_callback:
                raise

            if hasattr(ssl, 'ALERT_DESCRIPTION_BAD_CERTIFICATE'):
                return ssl.ALERT_DESCRIPTION_BAD_CERTIFICATE
            return True

        # Verification done.
        if as_callback:
            return None
        return True

    def disconnect(self):
        """ Disconnect from target. """
        if not self.connected:
            return

        with self.socket_lock:
            if self.tls:
                self.teardown_tls()

            # This might give an error if the connection was already closed by the other end.
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
            except OSError:
                pass
            self.socket.close()
            self.socket = None

        # Clear buffers and queues.
        with self.buffer_lock:
            self.buffer = None
        with self.message_lock:
            self.message_queue = None

    def teardown_tls(self):
        """ Tear down our TLS connection and give us our regular socket back. """
        # This might give an error if the connection was already closed by the other end.
        try:
            self.socket = self.socket.unwrap()
        except OSError:
            pass


    @property
    def connected(self):
        """ Whether this connection is... connected to something. """
        return self.socket is not None


    ## High-level message-related methods.

    def generate_messages(self):
        """ Create an iterator out of this connection which will blockingly generate messages as they come. """
        return iter(self)

    def has_message(self, types=None):
        """ Determine if this connection has a message ready for processing. """
        # Low-hanging fruit: is there a message ready in the queue?
        if len(self.message_queue) > 0:
            # Is the message of the type we want?
            if types:
                with self.message_lock:
                    for source, command, params in reversed(self.message_queue):
                        if command in types:
                            return True
                    else:
                        return False
            # Any message is fine.
            return True

        # See if we have data that isn't parsed yet.
        if self.has_data() and self.receive_data():
            added = self.parse_data()

            # Run again if we're looking for specific types, else it's all good.
            if added and types:
                return self.has_message(types=types)
            # Any added message is fine.
            return added

        return False

    def get_message(self, types=None, retry=False):
        """ Get an IRC message for processing. """
        with self.message_lock:
            try:
                # Looking for specific type?
                if types:
                    # Iterate through messages and find relevant type.
                    for source, command, params in reversed(self.message_queue[:]):
                        if command in types:
                            self.message_queue.remove((source, command, params))
                            return source, command, params
                    else:
                        raise IndexError('Not in queue.')
                # Any type is fine.
                else:
                    message = self.message_queue.popleft()
            except IndexError:
                if retry or not self.has_data():
                    raise NoMessageAvailable('No message available.')

                # Try to parse any messages we have and try again.
                if not self.receive_data() or not self.parse_data():
                    raise NoMessageAvailable('No message available.')

                # New messages were added, let's fetch them.
                return self.get_message(types=types, retry=True)

        return message

    def send_message(self, command, *params, source=None):
         """ Send a message to the other endpoint. """
         message = protocol.construct(command, *params, source=source)
         return self.send_string(message)

    def wait_for_message(self, types=None):
        """ Wait until a message has arrived. """
        # No use waiting if we already have a message.
        if self.has_message(types=types):
            return

        if not self.connected:
            raise NotConnected('Not connected.')

        while True:
            try:
                select.select([ self.socket ], [], [])
            except Exception as e:
                self.disconnect()
                raise NotConnected('Error while select()ing on socket: ' + str(e))

            if self.has_message(types=types):
                break


    ## Iterator stuff.

    def __iter__(self):
        """ Create an iterator out of this pool that will blockingly generate messages as they come. """
        return self

    def __next__(self):
        """ Wait for next message to arrive and return it. """
        if not self.connected:
            raise StopIteration

        self.wait_for_message()
        return self.get_message()


    ## Lower-level data-related methods.

    def has_data(self):
        """ Check if socket has data. """
        if not self.connected:
            return False

        with self.socket_lock:
            try:
                readable, writable, error = select.select([ self.socket ], [], [], 0)
            except Exception as e:
                self.disconnect()
                raise NotConnected('Error while select()ing on socket: ' + str(e))

            return self.socket in readable


    def receive_data(self):
        """ Try and receive any data and process it. Will return True if new data was added. """
        if not self.has_data():
            return False

        if hasattr(socket, 'MSG_NOSIGNAL'):
            flags = socket.MSG_NOSIGNAL
        else:
            flags = 0

        data = b''
        with self.socket_lock:
            try:
                data = self.socket.recv(BUFFER_SIZE, flags)
                # No data while select() indicates we have data available means the other party has closed the socket.
                if not data:
                    self.disconnect()
            except OSError as e:
                 if hasattr(errno, 'EAGAIN') and e.errno == errno.EAGAIN:
                     return self.receive_data()
                 raise

        if data:
            with self.buffer_lock:
                self.buffer += data
            return True
        return False

    def extract_data(self):
        """ Extract lines from current data. """
        # Some IRC servers use only \n as the line separator, so use that.
        sep = protocol.MINIMAL_LINE_SEPARATOR.encode('us-ascii')

        # Extract message lines from buffer.
        with self.buffer_lock:
            if not sep in self.buffer:
                return False

            lines = [ line + sep for line in self.buffer.split(sep) ]
            # Put last line (the remainder) back into the buffer, minus the added separator.
            self.buffer = lines.pop()[:-len(sep)]

        return lines

    def parse_data(self):
        """ Attempt to parse existing data into IRC messages. Will return True if new messages were added. """
        lines = self.extract_data()

        # Parse messages and put them into the message queue.
        with self.message_lock:
            for line in lines:
                # Parse message.
                try:
                    parsed = protocol.parse(line, encoding=self.encoding)
                except protocol.ProtocolViolation:
                    # TODO: Notification.
                    continue

                self.message_queue.append(parsed)

        return True

    def send_string(self, string):
        """ Send raw string to other end point. """
        self.send_data(string.encode(self.encoding))

    def send_data(self, data):
        """ Send raw data to other end point. """
        sent = 0

        if not self.connected:
            raise NotConnected('Not connected.')

        with self.socket_lock:
            # Continue until all data has been sent.
            while sent < len(data):
                try:
                    sent += self.socket.send(data[sent:])
                except OSError as e:
                    if hasattr(errno, 'EAGAIN') and e.errno == errno.EAGAIN:
                        continue
                    raise


class ConnectionPool:
    """ A pool of connections. """
    def __init__(self, conns=None):
        if not conns:
            conns = []

        self.connections = set(conns)
        self.connection_cycle = itertools.cycle(self.connections)
        self.index = 0

    def add(self, connection):
        """ Add connection to pool. """
        self.connections.add(connection)
        self.connection_cycle = itertools.cycle(self.connections)

    def remove(self, connection):
        self.connections.remove(connection)
        self.connection_cycle = itertools.cycle(self.connections)


    ## High-level message stuff.

    def generate_messages(self):
        """ Create an iterator out of this pool which will blockingly generate messages as they come. """
        return iter(self)

    def has_message(self):
        """ Check if any connection has messages available. """
        return any(conn.has_message() for conn in self.connections)

    def get_message(self):
        """
        Get first available message from any connection. Returns a (connection, message) tuple.
        Tries to be fair towards connections by cycling the start connection it tries to take a message from.
        """
        if not self.has_message():
            raise NoMessageAvailable('No message available.')

        for conn in self.connection_cycle:
            if conn.has_message():
                return (conn, conn.get_message())

    def wait_for_message(self):
        """ Wait until any connection has a message available. """
        sockets = { conn.socket: conn for conn in self.connections if conn.connected }
        found = False

        if self.has_message():
            return

        while not found:
            # Wait forever until a socket becomes readable.
            readable, writable, error = select.select(sockets.keys(), [], [])

            # Iterate and parse.
            for socket in readable:
                conn = sockets[socket]

                if conn.has_message():
                    found = True
                    break


    ## Lower-level data stuff.

    def has_data(self):
        """ Check if any socket has data available. """
        sockets = [ conn.socket for conn in self.connections if conn.connected ]

        readable, writable, error = select.select(sockets, [], [], 0)
        return len(readable) > 0


    ## Iterator stuff.

    def __iter__(self):
        """ Create an iterator out of this pool that will blockingly generate messages as they come. """
        return self

    def __next__(self):
        """ Wait for next message to arrive and return it. """
        self.wait_for_message()
        return self.get_message()




## nickserv.py
# Support automatic NickServ authentication.
from .. import client

__all__ = [ 'NickServSupport' ]


class NickServSupport(client.BasicClient):
    """ Support for NickServ authentication. """

    ## Internal overrides.

    def __init__(self, *args, nickserv_password=None, **kwargs):
        self.__init__(*args, **kwargs)
        self.nickserv_password = nickserv_password




## _args.py
# Common argument parsing code.
import argparse
import pydle

def client_from_args(name, description, default_nick='Bot', cls=pydle.Client):
    # Parse some arguments.
    parser = argparse.ArgumentParser(name, description=description, add_help=False,
        epilog='This program is part of {package}.'.format(package=pydle.__name__))

    meta = parser.add_argument_group('Meta')
    meta.add_argument('-h', '--help', action='help', help='What you are reading right now.')
    meta.add_argument('-v', '--version', action='version', version='{package}/%(prog)s {ver}'.format(package=pydle.__name__, ver=pydle.__version__), help='Dump version number.')
    meta.add_argument('-V', '--verbose', help='Be verbose in warnings and errors.', action='store_true', default=False)
    meta.add_argument('-d', '--debug', help='Show debug output.', action='store_true', default=False)

    conn = parser.add_argument_group('Connection')
    conn.add_argument('server', help='The server to connect to.', metavar='SERVER')
    conn.add_argument('-p', '--port', help='The port to use. (default: 6667, 6697 (TLS))')
    conn.add_argument('-P', '--password', help='Server password.', metavar='PASS')
    conn.add_argument('--tls', help='Use TLS. (default: no)', action='store_true', default=False)
    conn.add_argument('--verify-tls', help='Verify TLS certificate sent by server. (default: no)', action='store_true', default=False)
    conn.add_argument('-e', '--encoding', help='Connection encoding. (default: UTF-8)', default='utf-8', metavar='ENCODING')

    init = parser.add_argument_group('Initialization')
    init.add_argument('-n', '--nickname', help='Nickname. Can be set multiple times to set fallback nicknames. (default: {})'.format(default_nick), action='append', dest='nicknames', default=[], metavar='NICK')
    init.add_argument('-u', '--username', help='Username. (default: derived from nickname)', metavar='USER')
    init.add_argument('-r', '--realname', help='Realname (GECOS). (default: derived from nickname)', metavar='REAL')
    init.add_argument('-c', '--channel', help='Channel to automatically join. Can be set multiple times for multiple channels.', action='append', dest='channels', default=[], metavar='CHANNEL')

    auth = parser.add_argument_group('Authentication')
    auth.add_argument('--sasl-identity', help='Identity to use for SASL authentication. (default: <empty>)', default='', metavar='SASLIDENT')
    auth.add_argument('--sasl-username', help='Username to use for SASL authentication.', metavar='SASLUSER')
    auth.add_argument('--sasl-password', help='Password to use for SASL authentication.', metavar='SASLPASS')
    auth.add_argument('--tls-client-cert', help='TLS client certificate to use.', metavar='CERT')
    auth.add_argument('--tls-client-cert-keyfile', help='Keyfile to use for TLS client cert.', metavar='KEYFILE')
    auth.add_argument('--nickserv-password', help='NickServ password to identify with.', metavar='NSPASS')

    args = parser.parse_args()

    # Set nicknames straight.
    if not args.nicknames:
        nick = default_nick
        fallback = []
    else:
        nick = args.nicknames.pop(0)
        fallback = args.nicknames

    # Set log level.
    if args.debug:
        pydle.log.Logger.LEVEL = pydle.log.DEBUG
    elif not args.verbose:
        pydle.log.Logger.LEVEL = pydle.log.ERROR

    # Setup client and connect.
    client = cls(nickname=nick, fallback_nicknames=fallback, username=args.username, realname=args.realname,
        sasl_identity=args.sasl_identity, sasl_username=args.sasl_username, sasl_password=args.sasl_password,
        tls_client_cert=args.tls_client_cert, tls_client_cert_key=args.tls_client_cert_keyfile,
        nickserv_password=args.nickserv_password)

    client.connect(hostname=args.server, port=args.port, password=args.password, encoding=args.encoding,
        channels=args.channels, tls=args.tls, tls_verify=args.verify_tls)

    return client



## cap.py
# Server <-> client optional extension indication support.
# See also: http://ircv3.atheme.org/specification/capability-negotiation-3.1
import re
from .. import client

__all__ = [ 'CapabilityNegotiationSupport', 'NEGOTIATED', 'NEGOTIATING', 'FAILED' ]


DISABLED_PREFIX = '-'
ACKNOWLEDGEMENT_REQUIRED_PREFIX = '~'
STICKY_PREFIX = '='
PREFIXES = '-~='
NEGOTIATING = True
NEGOTIATED = None
FAILED = False


class CapabilityNegotiationSupport(client.BasicClient):
    """ CAP command support. """

    ## Internal overrides.

    def _reset_attributes(self):
        super()._reset_attributes()
        self._capabilities = {}
        self._capabilities_requested = set()
        self._capabilities_negotiating = set()

    def _register(self):
        """ Hijack registration to send a CAP LS first. """
        if self.registered:
           return

        # Ask server to list capabilities.
        self.rawmsg('CAP', 'LS')

        # Register as usual.
        super()._register()

    def _capability_normalize(self, cap):
        return cap.lstrip(PREFIXES).lower()


    ## API.

    def capability_negotiated(self, capab):
        """ Mark capability as negotiated, and end negotiation if we're done. """
        self._capabilities_negotiating.discard(capab)

        if not self._capabilities_requested and not self._capabilities_negotiating:
            self.rawmsg('CAP', 'END')


    ## Message handlers.

    def on_raw_cap(self, source, params):
        """ Handle CAP message. """
        target, subcommand, params = params[0], params[1], params[2:]

        # Call handler.
        attr = 'on_raw_cap_' + subcommand.lower()
        if hasattr(self, attr):
            getattr(self, attr)(params)
        else:
            self.logger.warning('Unknown CAP subcommand sent from server: {}', subcommand)

    def on_raw_cap_ls(self, params):
        """ Update capability mapping. Request capabilities. """
        to_request = set()

        for capab in params[0].split():
            cp = self._capability_normalize(capab)

            # Only process new capabilities.
            if cp in self._capabilities:
                continue

            # Check if we support the capability.
            attr = 'on_capability_' + capability_to_identifier(cp) + '_available'
            supported = getattr(self, attr)() if hasattr(self, attr) else False

            if supported:
                to_request.add(cp)
            else:
                self._capabilities[cp] = False

        if to_request:
            # Request some capabilities.
            self._capabilities_requested.union(to_request)
            self.rawmsg('CAP', 'REQ', ' '.join(to_request))
        else:
            # No capabilities requested, end negotiation.
            self.rawmsg('CAP', 'END')

    def on_raw_cap_list(self, params):
        """ Update active capabilities. """
        self._capabilities = { capab: False for capab in self._capabilities }

        for capab in params[0].split():
            capab = self._capability_normalize(capab)
            self._capabilities[capab] = True

    def on_raw_cap_ack(self, params):
        """ Update active capabilities: requested capability accepted. """
        for capab in params[0].split():
            cp = self._capability_normalize(capab)
            self._capabilities_requested.discard(cp)

            # Determine capability type and callback.
            if capab.startswith(DISABLED_PREFIX):
                self._capabilities[cp] = False
                attr = 'on_capability_' + capability_to_identifier(cp) + '_disabled'
            elif capab.startswith(STICKY_PREFIX):
                # Can't disable it. Do nothing.
                self.logger.err('Could not disable capability {}.', cp)
                continue
            else:
                self._capabilities[cp] = True
                attr = 'on_capability_' + capability_to_identifier(cp) + '_enabled'

            # Indicate we're gonna use this capability if needed.
            if capab.startswith(ACKNOWLEDGEMENT_REQUIRED_PREFIX):
                self.rawmsg('CAP', 'ACK', cp)

            # Run callback.
            if hasattr(self, attr):
                status = getattr(self, attr)()
            else:
                status = NEGOTIATED

            # If the process needs more time, add it to the database and end later.
            if status == NEGOTIATING:
                self._capabilities_negotiating.add(cp)
            elif status == FAILED:
                # Ruh-roh, negotiation failed. Disable the capability.
                self.logger.warn('Capability negotiation for {} failed. Attempting to disable capability again.', cp)

                self.rawmsg('CAP', 'REQ', '-' + cp)
                self._capabilities_requested.add(cp)

        # If we have no capabilities left to process, end it.
        if not self._capabilities_requested and not self._capabilities_negotiating:
            self.rawmsg('CAP', 'END')

    def on_raw_cap_nak(self, params):
        """ Update active capabilities: requested capability rejected. """
        for capab in params[0].split():
            capab = self._capability_normalize(capab)
            self._capabilities[capab] = False
            self._capabilities_requested.discard(capab)

        # If we have no capabilities left to process, end it.
        if not self._capabilities_requested and not self._capabilities_negoatiating:
            self.rawmsg('CAP', 'END')


    def on_raw_410(self, source, params):
        """ Unknown CAP subcommand or CAP error. Force-end negotiations. """
        self.logger.err('Server sent "Unknown CAP subcommand: {}". Aborting capability negotiation.', params[0])

        self._capabilities_requested = set()
        self._capabilities_negotiating = set()
        self.rawmsg('CAP', 'END')

    def on_raw_421(self, source, params):
        """ Hijack to ignore the absence of a CAP command. """
        if params[0] == 'CAP':
            return
        super().on_raw_421(source, params)

    def on_raw_451(self, source, params):
        """ Hijack to ignore the absence of a CAP command. """
        if params[0] == 'CAP':
            return
        super().on_raw_451(source, params)


## Helpers.

def capability_to_identifier(name):
    """ Clean up capability so it works for a Python identifier. """
    name = name.lower()
    name = re.sub('[^a-z]', '_', name)
    return name



## client.py
# Basic IRC client implementation.
import os
import datetime
import time
import itertools

from . import connection
from . import protocol
from . import log

__all__ = [ 'AlreadyInChannel', 'NotInChannel', 'BasicClient' ]


class IRCError(Exception):
    """ Base class for all pydle errors. """
    pass

class NotInChannel(IRCError):
    pass

class AlreadyInChannel(IRCError):
    pass


class BasicClient:
    """ Basic RFC1459 IRC client. """
    DEFAULT_QUIT_MESSAGE = 'Quitting'

    RECONNECT = True
    RECONNECT_MAX_ATTEMPTS = 3
    RECONNECT_DELAYED = True
    RECONNECT_DELAYS = [0, 30, 120, 600]

    def __init__(self, nickname, fallback_nicknames=[], username=None, realname=None, **kwargs):
        """ Create a client. """
        self._nicknames = [nickname] + fallback_nicknames
        self.username = username or nickname.lower()
        self.realname = realname or nickname
        self.logger = log.Logger()
        self._reset_connection_attributes()
        self._reset_attributes()

        if kwargs:
            self.logger.warn('Unused arguments: {kw}', kw=kwargs)

    def _reset_attributes(self):
        """ Reset attributes. """
        ## Internal variables.

        # Connection and basic state.
        self._registration_attempts = 0
        self._attempt_nicknames = self._nicknames[:]
        self._mode = {}

        # Record-keeping.
        self.channels = {}
        self.users = {}

        # Capabilities and features.
        self._extban_types = []

        # Limits.
        self._away_message_length_limit = None
        self._channel_length_limit = protocol.CHANNEL_LENGTH_LIMIT
        self._channel_limit_groups = protocol.CHANNEL_LIMITS_GROUPS.copy()
        self._channel_limits = protocol.CHANNEL_LIMITS.copy()
        self._command_parameter_limit = protocol.PARAMETER_LIMIT
        self._list_limit_groups = {}
        self._list_limits = {}
        self._mode_limit = None
        self._nickname_length_limit = protocol.NICKNAME_LENGTH_LIMIT
        self._target_limits = {}
        self._topic_length_limit = protocol.TOPIC_LENGTH_LIMIT

        # Modes and prefixes.
        self._channel_modes = set(protocol.CHANNEL_MODES)
        self._channel_modes_behaviour = protocol.CHANNEL_MODES_BEHAVIOUR.copy()
        self._channel_prefixes = set(protocol.CHANNEL_PREFIXES)
        self._extban_prefix = None
        self._nickname_prefixes = protocol.NICKNAME_PREFIXES.copy()
        self._status_message_prefixes = set()
        self._user_modes = set(protocol.USER_MODES)

        # Misc.
        self._case_mapping = protocol.CASE_MAPPING
        self.logger.name = self.__class__.__name__

        ## Public connection attributes.
        self.nickname = '<unregistered>'
        self.registered = False
        self.motd = None
        self.network = None

    def _reset_connection_attributes(self):
        """ Reset connection attributes. """
        self.connection = None
        self.password = None
        self._autojoin_channels = []
        self._has_quit = False
        self._reconnect_attempts = 0


    ## Connection.

    def connect(self, hostname, port=None, password=None, channels=[], encoding='utf-8'):
        """ Connect to IRC server. """
        # Disconnect from current connection.
        if self.connected:
            self.disconnect()
            self._reset_connection_attributes()

        self.password = password
        self._autojoin_channels = channels

        # Create connection.
        self.connection = connection.Connection(hostname, port or protocol.DEFAULT_PORT, encoding=encoding)
        # Connect.
        self.connection.connect()

        # And initiate the IRC connection.
        self._register()

    def disconnect(self):
        """ Disconnect from server. """
        if self.connected:
            self.connection.disconnect()
            self.on_disconnect()

            # Reset any attributes.
            self._reset_attributes()

    def reconnect(self):
        """ Reconnect to server. """
        if self.connected:
            self.disconnect()

        # Reuse connection.
        self.connection.connect()

        # Set logger name.
        self.logger.name = self.__class__.__name__ + ':' + self.server_tag

        # Initiate registration.
        self._register()

    def _register(self):
        """ Perform IRC connection registration. """
        if self.registered:
            return

        self._registration_attempts += 1

        # Password first.
        if self.password:
            self.rawmsg('PASS', self.password)

        # Then nickname...
        self.nickname = self._attempt_nicknames.pop(0)

        # And now for the rest of the user information.
        self.rawmsg('USER', self.username, '0', '*', self.realname)

    def _registration_completed(self, source, params):
        """ We're connected and registered. Receive proper nickname and emit fake NICK message. """
        self.registered = True

        target = params[0]
        self.on_raw_nick(self.nickname, [ target ])

    def _reconnect_delay(self):
        """ Calculate reconnection delay. """
        if self.RECONNECT and self.RECONNECT_DELAYED:
            if self._reconnect_attempts > len(self.RECONNECT_DELAYS):
                return self.RECONNECT_DELAYS[-1]
            else:
                return self.RECONNECT_DELAYS[self._reconnect_attempts]
        else:
            return 0


    ## Internal database management.

    def _create_channel(self, channel):
        self.channels[channel] = {
            'topic': None,
            'topic_by': None,
            'topic_set': None,
            'users': set(),
            'modes': {},
            'created': None,
            'password': None,
            'banlist': None,
            'public': True
        }

    def _destroy_channel(self, channel):
        del self.channels[channel]


    def _create_user(self, nickname):
        self.users[nickname] = {
            'away': False,
            'away_message': None,
            'username': None,
            'realname': None,
            'hostname': None
        }

    def _sync_user(self, nick, user, host):
        # Servers are NOT users.
        if '.' in nick:
            return

        # Create user in database.
        if not nick in self.users:
            self._create_user(nick)

        # Update user/host combination.
        if user:
            self.users[nick]['username'] = user
        if host:
            self.users[nick]['hostname'] = host

    def _rename_user(self, user, new):
        if user in self.users:
            self.users[new] = self.users[user]
            del self.users[user]
        else:
            self._create_user(new)

        for ch in self.channels.values():
            # Rename user in channel list.
            if user in ch['users']:
                ch['users'].discard(user)
                ch['users'].add(new)

            # Also replace in channel status lists.
            for mode in self._nickname_prefixes.values():
                if mode in ch['modes'] and user in ch['modes'][mode]:
                    ch['modes'][mode].remove(user)
                    ch['modes'][mode].append(new)

    def _destroy_user(self, nickname, channel=None):
        if channel:
            channels = [ self.channels[channel] ]
        else:
            channels = self.channels.values()

        for ch in channels:
            # Remove from nicklist.
            ch['users'].discard(nickname)

            # Remove from statuses.
            for status in self._nickname_prefixes.values():
                if status in ch['modes'] and nickname in ch['modes'][status]:
                    ch['modes'][status].remove(nickname)

        # If we're not in any common channels with the user anymore, we have no reliable way to keep their info up-to-date.
        # Remove the user.
        if not channel or not any(nickname in ch['users'] for ch in self.channels.values()):
            del self.users[nickname]


    ## IRC attributes.

    @property
    def connected(self):
        """ Whether or not we are connected. """
        return self.connection and self.connection.connected

    @property
    def nickname(self):
        """ Our currently active nickname. """
        return self._nickname

    @nickname.setter
    def nickname(self, value):
        """ Change nickname. """
        if self.connected:
            self.rawmsg('NICK', value)
        else:
            self._nickname = value

    @property
    def server_tag(self):
        if self.connected and self.connection.hostname:
            if self.network:
                tag = self.network.lower()
            else:
                tag = self.connection.hostname.lower()

                # Remove hostname prefix.
                if tag.startswith('irc.'):
                    tag = tag[4:]

                # Check if host is either an FQDN or IPv4.
                if '.' in tag:
                    # Attempt to cut off TLD.
                    host, suffix = tag.rsplit('.', 1)

                    # Make sure we aren't cutting off the last octet of an IPv4.
                    try:
                        int(suffix)
                    except ValueError:
                        tag = host

            return tag
        else:
            return None


    ## IRC API.

    def raw(self, message):
        """ Send raw command. """
        if not message.endswith(protocol.LINE_SEPARATOR):
            message += protocol.LINE_SEPARATOR
        self._send_raw(message)

    def rawmsg(self, command, *params, source=None):
        """ Send raw message. """
        self._send_message(command, *params, source=source)

    def join(self, channel, password=None):
        """ Join channel, optionally with password. """
        if self.in_channel(channel):
            raise AlreadyInChannel('Already in channel {}'.format(channel))

        if password:
            self.rawmsg('JOIN', channel, password)
        else:
            self.rawmsg('JOIN', channel)

    def part(self, channel, message=None):
        """ Leave channel, optionally with message. """
        if not self.in_channel(channel):
            raise NotInChannel('Not in channel {}'.format(channel))

        # Message seems to be an extension to the spec.
        if message:
            self.rawmsg('PART', channel, message)
        else:
            self.rawmsg('PART', channel)

    def quit(self, message=DEFAULT_QUIT_MESSAGE):
        """ Quit network. """
        self._has_quit = True
        self.rawmsg('QUIT', message)
        self.disconnect()

    def cycle(self, channel):
        """ Rejoin channel. """
        if not self.in_channel(channel):
            raise NotInChannel('Not in channel {}'.format(channel))

        password = self.channels[channel]['password']
        self.part(channel)
        self.join(channel, password)

    def message(self, target, message):
        """ Message channel or user. """
        if self.is_channel(target) and not self.in_channel(target):
            raise NotInChannel('Not in channel {}'.format(target))

        self.rawmsg('PRIVMSG', target, message)

    def notice(self, target, message):
        """ Notice channel or user. """
        if self.is_channel(target) and not self.in_channel(target):
            raise NotInChannel('Not in channel {}'.format(target))

        self.rawmsg('NOTICE', target, message)

    def mode(self, target, *modes):
        """ Set mode on target. """
        if self.is_channel(target) and not self.in_channel(target):
            raise NotInChannel('Not in channel {}'.format(target))

        self.rawmsg('MODE', target, *modes)

    def away(self, message):
        """ Mark self as away. """
        self.rawmsg('AWAY', message)

    def back(self):
        """ Mark self as not away. """
        self.rawmsg('AWAY')


    ## IRC helpers.

    def is_channel(self, chan):
        """ Check if given argument is a channel name or not. """
        return any(chan.startswith(prefix) for prefix in self._channel_prefixes)

    def in_channel(self, channel):
        return channel in self.channels.keys()

    def is_same_nick(self, left, right):
        """ Check if given nicknames are equal in the server's case mapping. """
        return protocol.equals(left, right, case_mapping=self._case_mapping)


    ## Overloadable callbacks.

    def on_connect(self):
        """ Callback called when the client has connected successfully. """
        # Reset reconnect attempts.
        self._reconnect_attempts = 0

        # Auto-join channels.
        for channel in self._autojoin_channels:
            self.join(channel)

    def on_disconnect(self):
        pass

    def on_invite(self, channel, by):
        pass

    def on_join(self, channel, user):
        pass

    def on_kill(self, target, by, reason):
        pass

    def on_kick(self, channel, target, by, reason=None):
        pass

    def on_mode_change(self, channel, modes, by):
        pass

    def on_user_mode_change(self, modes):
        pass

    def on_message(self, target, by, message):
        pass

    def on_channel_message(self, target, by, message):
        pass

    def on_private_message(self, by, message):
        pass

    def on_nick_change(self, old, new):
        pass

    def on_notice(self, target, by, message):
        pass

    def on_channel_notice(self, target, by, message):
        pass

    def on_private_notice(self, by, message):
        pass

    def on_part(self, channel, user, message=None):
        pass

    def on_topic_change(self, channel, message, by):
        pass

    def on_quit(self, user, message=None):
        pass


    ## Message dispatch.

    def _has_message(self, types=None):
        """ Whether or not we have messages available for processing. """
        return self.connection.has_message(types=types)

    def _wait_for_message(self, types=None):
        """ Wait for a received message. """
        self.connection.wait_for_message(types=types)

    def _get_message(self, types=None):
        """ Get a single message. """
        return self.connection.get_message(types=types)

    def _handle_message(self, types=None):
        """ Handle a single message. """
        source, command, params = self._get_message(types=types)
        self.logger.debug('<< [{source}] {command} {args}', source=source or '', command=command, args=params)

        if isinstance(command, int):
            cmd = str(command).zfill(3)
        else:
            cmd = command

        # Invoke dispatcher, if we have one.
        method = 'on_raw_' + cmd.lower()
        if hasattr(self, method):
            getattr(self, method)(source, params)
        # Invoke default method.
        else:
            self.on_unknown(command, source, params)

        return source, command, params

    def _send_raw(self, message):
        """ Send a raw message. """
        self.logger.debug('>>> {msg}', msg=message)
        self.connection.send_string(message)

    def _send_message(self, command, *params, source=None):
        """ Send a message. """
        self.logger.debug('>> [{source}] {command} {args}', source=source or '', command=command, args=params)
        self.connection.send_message(command, *params, source=source)

    def handle_forever(self):
        """ Handle messages forever. Main loop. """
        while True:
            while self.connected:
                try:
                    self._wait_for_message()
                    self._handle_message()
                except connection.NotConnected:
                    break

            if self._has_quit:
                # Nothing to worry about.
                break

            # Unexpected disconnect. Reconnect?
            if self.RECONNECT and (self.RECONNECT_MAX_ATTEMPTS is None or self._reconnect_attempts < self.RECONNECT_MAX_ATTEMPTS):
                # Calculate reconnect delay.
                delay = self._reconnect_delay()
                self._reconnect_attempts += 1

                if delay > 0:
                    self.logger.err('Unexpected disconnect. Attempting to reconnect within {n} seconds.', n=delay)
                else:
                    self.logger.err('Unexpected disconnect. Attempting to reconnect.')

                # Wait and reconnect.
                time.sleep(delay)
                self.reconnect()
            else:
                self.logger.err('Unexpected disconnect. Giving up.')


    ## Raw message handlers.

    def on_unknown(self, command, source, params):
        """ Unknown command. """
        self.logger.warn('Unknown command: [{source}] {command} {params}', source=source, command=command, params=params)

    def _ignored(self, source, params):
        """ Ignore message. """
        pass


    def on_raw_invite(self, source, params):
        """ INVITE command. """
        nick, user, host = protocol.parse_user(source)
        if nick in self.users:
            self._sync_user(nick, user, host)

        target, channel = params
        target = protocol.parse_user(target)

        if self.is_same_nick(self.nickname, nick):
            self.on_invite(channel, user)

    def on_raw_join(self, source, params):
        """ JOIN command. """
        nick, user, host = protocol.parse_user(source)
        self._sync_user(nick, user, host)

        channels = params[0].split(',')
        if self.is_same_nick(self.nickname, nick):
            # Add to our channel list, we joined here.
            for channel in channels:
                if not self.in_channel(channel):
                    self._create_channel(channel)

                # Request channel mode from IRCd.
                self.rawmsg('MODE', channel)
        else:
            # Add user to channel user list.
            for channel in channels:
                if self.in_channel(channel):
                    self.channels[channel]['users'].add(nick)

        for channel in channels:
            self.on_join(channel, nick)

    def on_raw_kick(self, source, params):
        """ KICK command. """
        kicker, kickeruser, kickerhost = protocol.parse_user(source)
        self._sync_user(kicker, kickeruser, kickerhost)

        if len(params) > 2:
            channels, targets, reason = params
        else:
            channels, targets = params
            reason = None

        channels = channels.split(',')
        targets = targets.split(',')

        for channel, target in zip(channels, targets):
            target, targetuser, targethost = protocol.parse_user(target)
            self._sync_user(target, targetuser, targethost)

            # Update nick list on channel.
            if self.in_channel(channel):
                self._destroy_user(target, channel)

            self.on_kick(channel, target, kicker, reason)

    def on_raw_kill(self, source, params):
        """ KILL command. """
        by, byuser, byhost = protocol.parse_user(source)
        target = protocol.parse_user(params[0])[0]
        message = params[1]

        if by in self.users:
            self._sync_user(by, byuser, byhost)

        if self.is_same_nick(self.nickname, target):
            self.disconnect()
        else:
            self._destroy_user(target)

        self.on_kill(target, by, reason)

    def on_raw_mode(self, source, params):
        """ MODE command. """
        nick, user, host = protocol.parse_user(source)
        target, modes = params[0], params[1:]

        self._sync_user(nick, user, host)

        if self.is_channel(target):
            if self.in_channel(target):
                # Parse modes.
                self.channels[target]['modes'] = protocol.parse_channel_modes(modes, self.channels[target]['modes'], behaviour=self._channel_modes_behaviour)

                self.on_mode_change(target, modes, nick)
        else:
            target, targetuser, targethost = protocol.parse_user(target)
            self._sync_user(target, targetuser, targethost)

            # Update own modes.
            if self.is_same_nick(self.nickname, nick):
                self._mode = protocol.parse_user_modes(modes, self._mode)

            self.on_user_mode_change(modes)

    def on_raw_nick(self, source, params):
        """ NICK command. """
        nick, user, host = protocol.parse_user(source)
        new = params[0]

        self._sync_user(nick, user, host)
        # Acknowledgement of nickname change: set it internally, too.
        # Alternatively, we were force nick-changed. Nothing much we can do about it.
        if self.is_same_nick(self.nickname, nick):
            self._nickname = new
            # Reflect logger change.
            self.logger.name = self.__class__.__name__ + ':' + self.server_tag + ':' + self._nickname

        # Go through all user lists and replace.
        self._rename_user(nick, new)

        self.on_nick_change(nick, new)

    def on_raw_notice(self, source, params):
        """ NOTICE command. """
        nick, user, host = protocol.parse_user(source)
        target, message = params

        if nick in self.users:
            self._sync_user(nick, user, host)

        self.on_notice(target, nick, message)
        if self.is_channel(target):
            self.on_channel_notice(target, nick, message)
        else:
            self.on_private_notice(nick, message)

    def on_raw_part(self, source, params):
        """ PART command. """
        nick, user, host = protocol.parse_user(source)
        channels = params[0].split(',')
        if len(params) > 1:
            message = params[1]
        else:
            message = None

        self._sync_user(nick, user, host)

        if self.is_same_nick(self.nickname, nick):
            # We left the channel. Remove from channel list. :(
            for channel in channels:
                if self.in_channel(channel):
                    del self.channels[channel]
        else:
            # Someone else left. Remove them.
            for ch in channels:
                self._destroy_user(nick, ch)

        for channel in channels:
            self.on_part(channel, nick, message)

    def on_raw_ping(self, source, params):
        """ PING command. """
        # Respond with a pong.
        self.rawmsg('PONG', *params)

    def on_raw_privmsg(self, source, params):
        """ PRIVMSG command. """
        nick, user, host = protocol.parse_user(source)
        target, message = params

        if nick in self.users:
            self._sync_user(nick, user, host)

        self.on_message(target, nick, message)
        if self.is_channel(target):
            self.on_channel_message(target, nick, message)
        else:
            self.on_private_message(nick, message)

    def on_raw_quit(self, source, params):
        """ QUIT command. """
        # No need to sync if the user is quitting anyway.
        user = protocol.parse_user(source)[0]
        if params:
            message = params[0]
        else:
            message = None

        # Remove user from database.
        if not self.is_same_nick(self.nickname, user):
            self._destroy_user(user)
        # Else, we quit.
        elif self.connected:
            self.disconnect()

        self.on_quit(user, message)

    def on_raw_topic(self, source, params):
        """ TOPIC command. """
        setter, setteruser, setterhost = protocol.parse_user(source)
        target, message = params

        self._sync_user(setter, setteruser, setterhost)

        # Update topic in our own channel list.
        if self.in_channel(target):
            self.channels[target]['topic'] = message
            self.channels[target]['topic_by'] = setter
            self.channels[target]['topic_set'] = datetime.datetime.now()

        self.on_topic_change(target, message, setter)


    ## Numeric responses.

    # Since RFC1459 specifies no specific banner message upon completion of registration,
    # take any of the below commands as an indication that registration succeeded.

    on_raw_001 = _registration_completed # Welcome message.
    on_raw_002 = _registration_completed # Server host.
    on_raw_003 = _registration_completed # Server creation time.

    def on_raw_004(self, source, params):
        """ Basic server information. """
        hostname, ircd, user_modes, channel_modes = params[:4]

        # Set valid channel and user modes.
        self._channel_modes = set(channel_modes)
        self._user_modes = set(user_modes)

    on_raw_008 = _registration_completed # Server notice mask.
    on_raw_042 = _registration_completed # Unique client ID.
    on_raw_250 = _registration_completed # Connection statistics.
    on_raw_251 = _registration_completed # Amount of users online.
    on_raw_252 = _registration_completed # Amount of operators online.
    on_raw_253 = _registration_completed # Amount of unknown connections.
    on_raw_254 = _registration_completed # Amount of channels.
    on_raw_255 = _registration_completed # Amount of local users and servers.
    on_raw_265 = _registration_completed # Amount of local users.
    on_raw_266 = _registration_completed # Amount of global users.

    def on_raw_324(self, source, params):
        """ Channel mode. """
        target, channel, modes = params[0], params[1], params[2:]
        if not self.in_channel(channel):
            return

        self.channels[channel]['modes'] = protocol.parse_channel_modes(modes, self.channels[channel]['modes'], behaviour=self._channel_modes_behaviour)

    def on_raw_329(self, source, params):
        """ Channel creation time. """
        target, channel, timestamp = params
        if not self.in_channel(channel):
            return

        self.channels[channel]['created'] = datetime.datetime.fromtimestamp(int(timestamp))

    def on_raw_332(self, source, params):
        """ Current topic on channel join. """
        target, channel, topic = params
        if not self.in_channel(channel):
            return

        self.channels[channel]['topic'] = topic

    def on_raw_333(self, source, params):
        """ Topic setter and time on channel join. """
        target, channel, setter, timestamp = params
        if not self.in_channel(channel):
            return

        # No need to sync user since this is most likely outdated info.
        self.channels[channel]['topic_by'] = protocol.parse_user(setter)[0]
        self.channels[channel]['topic_set'] = datetime.datetime.fromtimestamp(int(timestamp))

    def on_raw_353(self, source, params):
        """ Response to /NAMES. """
        target, visibility, channel, names = params
        if not self.in_channel(channel):
            return

        # Set channel visibility.
        if visibility == protocol.PUBLIC_CHANNEL_SIGIL:
            self.channels[channel]['public'] = True
        elif visibility in (protocol.PRIVATE_CHANNEL_SIGIL, protocol.SECRET_CHANNEL_SIGIL):
            self.channels[channel]['public'] = False

        # Update channel user list.
        for entry in names.split():
            statuses = []
            # Make entry safe for protocol.parse_user().
            safe_entry = entry.lstrip(''.join(self._nickname_prefixes.keys()))
            # Parse entry and update database.
            nick, user, host = protocol.parse_user(safe_entry)
            self._sync_user(nick, user, host)

            # Get prefixes.
            prefixes = set(entry.replace(safe_entry, ''))

            # Check, record and strip status prefixes.
            for prefix, status in self._nickname_prefixes.items():
                # Add to list of statuses by user.
                if prefix in prefixes:
                    statuses.append(status)

            # Add user to user list.
            self.channels[channel]['users'].add(nick)
            # And to channel modes..
            for status in statuses:
                if status not in self.channels[channel]['modes']:
                    self.channels[channel]['modes'][status] = []
                self.channels[channel]['modes'][status].append(nick)

    on_raw_366 = _ignored # End of /NAMES list.

    def on_raw_375(self, source, params):
        """ Start message of the day. """
        self._registration_completed(source, params)
        self.motd = params[1] + '\n'

    def on_raw_372(self, source, params):
        """ Append message of the day. """
        self.motd += params[1] + '\n'

    def on_raw_376(self, source, params):
        """ End of message of the day. """
        self.motd += params[1] + '\n'

        # MOTD is done, let's tell our bot the connection is ready.
        self.on_connect()

    def on_raw_421(self, source, params):
        """ Server responded with 'unknown command'. """
        self.logger.warn('Server responded with "Unknown command: {}"'.format(params[0]))

    def on_raw_433(self, source, params):
        """ Nickname in use. """
        if not self.registered:
            self._registration_attempts += 1
            # Attempt to set new nickname.
            if not self._attempt_nicknames:
                self.nickname = self._nicknames[0] + '_' * (self._registration_attempts - len(self._nicknames))
            else:
                self.nicknames = self._attempt_nicknames.pop(0)

    def on_raw_432(self, source, params):
        """ Erroneous nickname. """
        if not self.registered:
            self.on_raw_432(source, params)

    on_raw_436 = _ignored # Nickname collision, issued right before the server kills us.

    def on_raw_451(self, source, params):
        """ We have to register first before doing X. """
        # TODO: Implement. Warning?
        pass

    on_raw_451 = _ignored # You have to register first.
    on_raw_462 = _ignored # You may not re-register.


class ClientPool:
    """ A pool of clients. """

    def __init__(self, clients=None):
        if not clients:
            clients = []

        self.clients = set(clients)
        self.client_cycle = itertools.cycle(self.clients)
        self.connpool = connection.ConnectionPool(client.connection for client in self.clients)

    def add(self, client):
        """ Add client to pool. """
        self.clients.add(client)
        self.client_cycle = itertools.cycle(self.clients)
        self.connpool = connection.ConnectionPool(client.connection for client in self.clients)

    def remove(self, client):
        """ Remove client from pool. """
        self.clients.remove(client)
        self.client_cycle = itertools.cycle(self.clients)
        self.connpool = connection.ConnectionPool(client.connection for client in self.clients)


    ## High-level message stuff.

    def has_message(self):
        return self.connpool.has_message()

    def handle_message(self):
        """
        Handle first available message from any client in the pool.
        Tries to be fair towards clients by cycling the start client tries to take a message from.
        """
        if not self.has_message():
            raise connection.NoMessageAvailable('No message available.')

        for client in self.client_cycle:
            if client._has_message():
                return client._handle_message()

    def handle_forever(self):
        """ Enter infinite handle loop. """
        while True:
            self.wait_for_message()
            self.handle_message()

    def wait_for_message(self):
        return self.connpool.wait_for_message()




from . import ctcp, tls, isupport, cap, sasl, ircv3_1

from .ctcp import CTCPSupport
from .tls import TLSSupport
from .isupport import ISUPPORTSupport
from .cap import CapabilityNegotiationSupport
from .sasl import SASLSupport
from .ircv3_1 import IRCv3_1Support
from .ircv3_2 import IRCv3_2Support

ALL = [ CTCPSupport, IRCv3_1Support, IRCv3_2Support, ISUPPORTSupport ]
LITE = [ CTCPSupport, ISUPPORTSupport ]



#!/usr/bin/env python3
## console.py
# Interactive console.
import threading
import code

import pydle
from . import _args

def main():
    client = _args.client_from_args('ipydle', description='Interactive pydle console.')

    # Let client do stuff in background.
    thread = threading.Thread(target=client.handle_forever)
    thread.start()

    # Run console.
    interp = code.InteractiveConsole({ 'self': client })
    interp.interact('{name} {ver} interactive console'.format(name=pydle.__name__, ver=pydle.__version__))

    # Kill off.
    if client.connected:
        client.quit()
    thread.join()

if __name__ == '__main__':
    main()



# logging.py
# Thin wrapper around Python's (atrocious) logging module.
import logging

__all__ = [ 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'FATAL', 'Logger' ]


# Log levels.
FATAL = logging.FATAL
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG


class Logger:
    """
    Thin wrapper around Python's rather atrocious logging module in order to make it more bearable.
    Formats messages automatically according to advanced string formatting, handles logger name and logfile changes painlessly.
    """
    FORMAT = '{asctime} [{name}] {levelname}: {message}'
    DATE_FORMAT = None
    FILE = None
    LEVEL = logging.WARNING

    def __init__(self, name=None, file=None, formatter=None):
        self._file = file or self.FILE
        self._formatter = formatter or logging.Formatter(self.FORMAT, datefmt=self.DATE_FORMAT, style='{')
        if name:
            self._name = name
            self.refresh_logger()

    def refresh_logger(self):
        """ Recreate logger instance. """
        self.logger = logging.Logger(self.name)
        self.setup_logger()

    def setup_logger(self):
        """ Set logger settings. """
        self.logger.setLevel(self.LEVEL)

        handler = logging.StreamHandler()
        handler.setFormatter(self.formatter)
        self.logger.addHandler(handler)

        if self.file:
            handler = logging.FileHandler(self.file)
            handler.setFormatter(self.formatter)
            self.logger.addHandler(handler)


    @property
    def name(self):
        """ This logger's identifier. """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.refresh_logger()

    @property
    def file(self):
        """ The file we are logging to, if any. """
        return self._file

    @file.setter
    def file(self, value):
        self._file = value
        self.refresh_logger()

    @property
    def formatter(self):
        """ The formatter we are using. """
        return self._formatter

    @formatter.setter
    def formatter(self, value):
        self._formatter = value
        self.refresh_logger()


    def debug(self, message, *args, **kwargs):
        """ Log DEBUG message. """
        self.logger.debug(message.format(*args, **kwargs))

    def inform(self, message, *args, **kwargs):
        """ Log INFO message. """
        self.logger.info(message.format(*args, **kwargs))

    def warn(self, message, *args, **kwargs):
        """ Log WARNING message. """
        self.logger.warning(message.format(*args, **kwargs))

    def err(self, message, *args, **kwargs):
        """ Log ERROR message. """
        self.logger.error(message.format(*args, **kwargs))

    def exception(self, exception):
        """ Log exception. """
        self.logger.exception(exception)

    def fatal(self, message, *args, **kwargs):
        """ Log FATAL message. """
        self.logger.fatal(message.format(*args, **kwargs))




## ircv3_2.py
# IRCv3.2 support (in progress).
from .. import client
from .. import protocol

from . import isupport
from . import cap

__all__ = [ 'IRCv3_2Support' ]


class IRCv3_2Support(cap.CapabilityNegotiationSupport, isupport.ISUPPORTSupport):
    """ Support for some of IRCv3.2's extensions. Currently supported: chghost, userhost-in-names. """

    ## IRC callbacks.

    def on_capability_chghost_available(self):
        """ Server reply to indicate a user we are in a common channel with changed user and/or host. """
        return True

    def on_capability_userhost_in_names_available(self):
        """ Show full user!nick@host in NAMES list. We already parse it like that. """
        return True

    def on_capability_uhnames_available(self):
        """ Possibly outdated alias for userhost-in-names. """
        return self.on_capability_userhost_in_names_available()

    def on_isupport_uhnames(self, value):
        """ Let the server know that we support UHNAMES using the old ISUPPORT method, for legacy support. """
        self.rawmsg('PROTOCTL', 'UHNAMES')


    ## Message handlers.

    def on_raw_chghost(self, source, params):
        """ Change user and/or host of user. """
        if 'chghost' not in self._capabilities or not self._capabilities['chghost']:
            return

        nick = protocol.parse_user(source)[0]
        if nick not in self.users:
            return

        # Update user and host.
        user, host = params
        if self.is_same_nick(self.nickname, nick):
            self.username = user
            self.hostname = host
        else:
            self.users[nick]['username'] = user
            self.users[nick]['hostname'] = host



## isupport.py
# ISUPPORT (server-side IRC extension indication) support.
# See: http://tools.ietf.org/html/draft-hardy-irc-isupport-00
import collections

from .. import client
from .. import protocol

__all__ = [ 'ISUPPORTSupport' ]


FEATURE_DISABLED_PREFIX = '-'
BAN_EXCEPT_MODE = 'e'
INVITE_EXCEPT_MODE = 'I'


class ISUPPORTSupport(client.BasicClient):
    """ ISUPPORT support. """

    ## Internal overrides.

    def _reset_attributes(self):
        super()._reset_attributes()
        self._isupport = {}

    def _create_channel(self, channel):
        """ Create channel with optional ban and invite exception lists. """
        super()._create_channel(channel)
        if 'EXCEPTS' in self._isupport:
            self.channels[channel]['exceptlist'] = None
        if 'INVEX' in self._isupport:
            self.channels[channel]['inviteexceptlist'] = None

    ## Command handlers.

    def on_raw_005(self, source, params):
        """ ISUPPORT indication. """
        isupport = {}

        # Parse response.
        # Strip target (first argument) and 'are supported by this server' (last argument).
        for feature in params[1:-1]:
            if feature.startswith(FEATURE_DISABLED_PREFIX):
                value = False
            elif '=' in feature:
                feature, value = feature.split('=', 1)
            else:
                value = True
            isupport[feature.upper()] = value

        # Update internal dict first.
        self._isupport.update(isupport)

        # And have callbacks update other internals.
        for entry, value in isupport.items():
            if value != False:
                # A value of True technically means there was no value supplied; correct this for callbacks.
                if value == True:
                    value = None

                method = 'on_isupport_' + entry.lower()
                if hasattr(self, method):
                    getattr(self, method)(value)


    ## ISUPPORT handlers.

    def on_isupport_awaylen(self, value):
        """ Away message length limit. """
        self._away_message_length_limit = int(value)

    def on_isupport_casemapping(self, value):
        """ IRC case mapping for nickname comparisons. """
        if value in protocol.CASE_MAPPINGS:
            self._case_mapping = value

    def on_isupport_channellen(self, value):
        """ Channel name length limit. """
        self._channel_length_limit = int(value)

    def on_isupport_chanlimit(self, value):
        """ Simultaneous channel limits for user. """
        self._channel_limits = {}

        for entry in value.split(','):
            types, limit = entry.split(':')

            # Assign limit to channel type group and add lookup entry for type.
            self._channel_limits[frozenset(types)] = int(limit)
            for prefix in types:
                self._channel_limit_groups[prefix] = frozenset(types)

    def on_isupport_chanmodes(self, value):
        """ Valid channel modes and their behaviour. """
        list, param, param_set, noparams = [ set(modes) for modes in value.split(',')[:4] ]
        self._channel_modes.union(set(value.replace(',', '')))

        # The reason we have to do it like this is because other ISUPPORTs (e.g. PREFIX) may update these values as well.
        if not protocol.BEHAVIOUR_LIST in self._channel_modes_behaviour:
            self._channel_modes_behaviour[protocol.BEHAVIOUR_LIST] = set()
        self._channel_modes_behaviour[protocol.BEHAVIOUR_LIST].union(list)

        if not protocol.BEHAVIOUR_PARAMETER in self._channel_modes_behaviour:
            self._channel_modes_behaviour[protocol.BEHAVIOUR_PARAMETER] = set()
        self._channel_modes_behaviour[protocol.BEHAVIOUR_PARAMETER].union(param)

        if not protocol.BEHAVIOUR_PARAMETER_ON_SET in self._channel_modes_behaviour:
            self._channel_modes_behaviour[protocol.BEHAVIOUR_PARAMETER_ON_SET] = set()
        self._channel_modes_behaviour[protocol.BEHAVIOUR_PARAMETER_ON_SET].union(param_set)

        if not protocol.BEHAVIOUR_NO_PARAMETER in self._channel_modes_behaviour:
            self._channel_modes_behaviour[protocol.BEHAVIOUR_NO_PARAMETER] = set()
        self._channel_modes_behaviour[protocol.BEHAVIOUR_NO_PARAMETER].union(noparams)

    def on_isupport_chantypes(self, value):
        """ Channel name prefix symbols. """
        if not value:
            value = ''
        self._channel_prefixes = set(value)

    def on_isupport_excepts(self, value):
        """ Server allows ban exceptions. """
        if not value:
            value = BAN_EXCEPT_MODE
        self._channel_modes.add(value)
        self._channel_modes_behaviour['list'].add(value)

    def on_isupport_extban(self, value):
        """ Extended ban prefixes. """
        self._extban_prefix, types = value.split(',')
        self._extban_types = set(types)

    def on_isupport_invex(self, value):
        """ Server allows invite exceptions. """
        if not value:
            value = INVITE_EXCEPT_MODE
        self._channel_modes.add(value)
        self._channel_modes_behaviour['list'].add(value)

    def on_isupport_maxbans(self, value):
        """ Maximum entries in ban list. Replaced by MAXLIST. """
        if 'MAXLIST' not in self._isupport:
            if not self._list_limits:
                self._list_limits = {}
            self._list_limits['b'] = int(value)

    def on_isupport_maxchannels(self, value):
        """ Old version of CHANLIMIT. """
        if 'CHANTYPES' in self._isupport and 'CHANLIMIT' not in self._isupport:
            self._channel_limits = {}

            prefixes = self._isupport['CHANTYPES']
            # Assume the limit is for all types of channels. Make a single group for all types.
            self._channel_limits[frozenset(prefixes)] = int(value)
            for prefix in prefixes:
                self._channel_limit_groups[prefix] = frozenset(prefixes)

    def on_isupport_maxlist(self, value):
        """ Limits on channel modes involving lists. """
        self._list_limits = {}

        for entry in value.split(','):
            modes, limit = entry.split(':')

            # Assign limit to mode group and add lookup entry for mode.
            self._list_limits[frozenset(modes)] = int(limit)
            for mode in modes:
                self._list_limit_groups[mode] = frozenset(modes)

    def on_isupport_maxpara(self, value):
        """ Limits to parameters given to command. """
        self._command_parameter_limit = int(value)

    def on_isupport_modes(self, value):
        """ Maximum number of variable modes to change in a single MODE command. """
        self._mode_limit = int(value)

    def on_isupport_namesx(self, value):
        """ Let the server know we do in fact support NAMESX. Effectively the same as CAP multi-prefix. """
        self.rawmsg('PROTOCTL', 'NAMESX')

    def on_isupport_network(self, value):
        """ IRC network name. """
        self.network = value

    def on_isupport_nicklen(self, value):
        """ Nickname length limit. """
        self._nickname_length_limit = int(value)

    def on_isupport_prefix(self, value):
        """ Nickname prefixes on channels and their associated modes. """
        if not value:
            # No prefixes support.
            self._nickname_prefixes = collections.OrderedDict()
            return

        modes, prefixes = value.lstrip('(').split(')', 1)

        # Update valid channel modes and their behaviour as CHANMODES doesn't include PREFIX modes.
        self._channel_modes.union(set(prefixes))
        if not protocol.BEHAVIOUR_PARAMETER in self._channel_modes_behaviour:
            self._channel_modes_behaviour[protocol.BEHAVIOUR_PARAMETER] = set()
        self._channel_modes_behaviour[protocol.BEHAVIOUR_PARAMETER].union(set(prefixes))

        self._nickname_prefixes = collections.OrderedDict()
        for mode, prefix in zip(modes, prefixes):
            self._nickname_prefixes[prefix] = mode

    def on_isupport_statusmsg(self, value):
        """ Support for messaging every member on a channel with given status or higher. """
        self._status_message_prefixes.update(value)

    def on_isupport_targmax(self, value):
        """ The maximum number of targets certain types of commands can affect. """
        if not value:
            return

        for entry in value.split(','):
            command, limit = entry.split(':', 1)
            if not limit:
                continue
            self._target_limits[command] = int(limit)

    def on_isupport_topiclen(self, value):
        """ Channel topic length limit. """
        self._topic_length_limit = int(value)

    def on_isupport_wallchops(self, value):
        """ Support for messaging every opped member or higher on a channel. Replaced by STATUSMSG. """
        for prefix, mode in self._nickname_prefixes.items():
            if mode == 'o':
                break
        else:
            prefix = '@'
        self._status_message_prefixes.add(prefix)

    def on_isupport_wallvoices(self, value):
        """ Support for messaging every voiced member or higher on a channel. Replaced by STATUSMSG. """
        for prefix, mode in self._nickname_prefixes.items():
            if mode == 'v':
                break
        else:
            prefix = '+'
        self._status_message_prefixes.add(prefix)




## ircv3_1.py
# IRCv3.1 full spec support.
from . import sasl
from . import tls
from .. import protocol

__all__ = [ 'IRCv3_1Support' ]


NO_ACCOUNT = '*'

class IRCv3_1Support(sasl.SASLSupport, tls.TLSSupport):
    """ Support for IRCv3.1's base and optional extensions. """

    ## Internal overrides.

    def _create_user(self, nickname):
        super()._create_user(nickname)
        if ('account-notify' in self._capabilities and self._capabilities['account-notify']) or ('extended-join' in self._capabilities and self._capabilites['extended-join']):
            self.users[nickname]['account'] = None


    ## IRC callbacks.

    def on_capability_account_notify_available(self):
        """ Take note of user account changes. """
        return True

    def on_capability_away_notify_available(self):
        """ Take note of AWAY messages. """
        return True

    def on_capability_extended_join_available(self):
        """ Take note of user account and realname on JOIN. """
        return True

    def on_capability_multi_prefix_available(self):
        """ Thanks to how underlying client code works we already support multiple prefixes. """
        return True

    def on_capability_tls_available(self):
        """ We never need to request this explicitly. """
        return False


    ## Message handlers.

    def on_raw_account(self, source, params):
        """ Changes in the associated account for a nickname. """
        if 'account-notify' not in self._capabilities or not self._capabilities['account-notify']:
            return

        nick, user, host = protocol.parse_user(source)
        account = params[0]

        if nick not in self.users:
            return

        self._sync_user(nick, user, host)
        if account == NO_ACCOUNT:
            account = None
        self.users[nick]['account'] = account

    def on_raw_away(self, source, params):
        """ Process AWAY messages. """
        if 'away-notify' not in self._capabilities or not self._capabilities['away-notify']:
            return

        nick, user, host = protocol.parse_user(source)[0]
        if nick not in self.users:
            return

        self._sync_user(nick, user, host)
        self.users[nick]['away'] = len(params) > 0
        self.users[nick]['away_message'] = params[0] if len(params) > 0 else None

    def on_raw_join(self, source, params):
        """ Process extended JOIN messages. """
        if 'extended-join' in self._capabilities and self._capabilities['extended-join']:
            nick = protocol.parse_user(source)[0]
            channels, account, realname = params
            super().on_raw_join(source, [ channels ])

            if account == NO_ACCOUNT:
                account = None
            self.users[nick]['account'] = account
            self.users[nick]['realname'] = realname
        else:
            super().on_raw_join(source, params)



from . import connection, protocol, log, client, features

from .connection import NotConnected, NoMessageAvailable
from .client import NotInChannel, AlreadyInChannel, BasicClient, ClientPool
from .features.cap import NEGOTIATING as CAPABILITY_NEGOTIATING, FAILED as CAPABILITY_FAILED, NEGOTIATED as CAPABILITY_NEGOTIATED

__name__ = 'pydle'
__version__ = '0.7.0'
__version_info__ = (0, 7, 0)
__license__ = 'BSD'


def featurize(*features):
    """ Put features into proper MRO order. """
    from functools import cmp_to_key

    def compare_subclass(left, right):
        if issubclass(left, right):
            return -1
        elif issubclass(right, left):
            return 1
        return 0

    sorted_features = sorted(features, key=cmp_to_key(compare_subclass))
    name = 'FeaturizedClient[{features}]'.format(features=', '.join(feature.__name__ for feature in sorted_features))
    return type(name, tuple(sorted_features), {})

class Client(featurize(*features.ALL)):
    """ A fully featured IRC client. """
    pass

class MinimalClient(featurize(*features.LITE)):
    """ A cut-down, less-featured IRC client. """
    pass



#!/usr/bin/env python3
## irccat.py
# Simple threaded irccat implementation, using pydle.
import sys
import threading

from .. import features
from .. import log
from .. import protocol

from .. import featurize, __version__
from . import _args


class IRCCat(featurize(*features.ALL)):
    """ irccat. Takes raw messages on stdin, dumps raw messages to stdout. Life has never been easier. """
    def _get_message(self, types=None):
        """ Get message and print it to stdout. """
        source, command, params = self.connection.get_message(types=types)
        sys.stdout.write(protocol.construct(command, *params, source=source))
        return source, command, params

    def _send_message(self, command, *params, source=None):
        sys.stdout.write(protocol.construct(command, *params, source=source))
        super()._send_message(command, *params, source=source)

    def _send_raw(self, raw):
        sys.stdout.write(raw)
        super()._send_raw(raw)

    def process_stdin_forever(self):
        """ Yes. """
        while True:
            line = sys.stdin.readline()
            if not line:
                break
            self.raw(line)

    def on_ctcp_version(self, source, target):
        self.ctcp_reply(source, 'VERSION', 'pydle-irccat v{}'.format(__version__))


def main():
    log.Logger.FORMAT = '!! {levelname}: {message}'

    # Create client.
    irccat = _args.client_from_args('irccat', default_nick='irccat', description='Process raw IRC messages from stdin, dump received IRC messages to stdout.', cls=IRCCat)

    thread = None
    if irccat.connected:
        # Let's rock.
        thread = threading.Thread(target=irccat.handle_forever)
        thread.start()

        # Process input in main thread.
        irccat.process_stdin_forever()

    # Other thread is done. Let's disconnect.
    irccat.quit('EOF on standard input')
    if thread:
        thread.join()

if __name__ == '__main__':
    main()



## sasl.py
# SASL authentication support. Currently we only support PLAIN authentication.
import threading
import base64
try:
    import puresasl
    import puresasl.client
except ImportError:
    puresasl = None

from . import cap

__all__ = [ 'SASLSupport' ]


RESPONSE_LIMIT = 400
EMPTY_MESSAGE = '+'
ABORT_MESSAGE = '*'


class SASLSupport(cap.CapabilityNegotiationSupport):
    """ SASL authentication support. Currently limited to the PLAIN mechanism. """
    SASL_TIMEOUT = 10

    ## Internal overrides.

    def __init__(self, *args, sasl_identity='', sasl_username=None, sasl_password=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.sasl_identity = sasl_identity
        self.sasl_username = sasl_username
        self.sasl_password = sasl_password

    def _reset_attributes(self):
        super()._reset_attributes()
        self._sasl_aborted = False
        self._sasl_challenge = b''
        self._sasl_timer = None


    ## SASL functionality.

    def _sasl_start(self):
        """ Initiate SASL authentication. """
        # The rest will be handled in on_raw_authenticate()/_sasl_respond().
        self.rawmsg('AUTHENTICATE', 'PLAIN')

    def _sasl_abort(self, timeout=False):
        """ Abort SASL authentication. """
        # You Only Abort Once
        if not self._sasl_aborted:
            if timeout:
                self.logger.err('SASL authentication timed out: aborting.')
            else:
                self.logger.err('SASL authentication aborted.')
            self._sasl_aborted = True

            # We're done here.
            self.rawmsg('AUTHENTICATE', ABORT_MESSAGE)
            self.capability_negotiated('sasl')

    def _sasl_end(self):
        """ Finalize SASL authentication. """
        self.capability_negotiated('sasl')

    def _sasl_respond(self):
        """ Respond to SASL challenge with response. """
        # Formulate a response.
        saslclient = puresasl.client.SASLClient(self.connection.hostname, 'irc', mechanism='PLAIN',
            username=self.sasl_username, password=self.sasl_password, identity=self.sasl_identity)

        response = base64.b64encode(saslclient.process(challenge=self._sasl_challenge)).decode(self.connection.encoding)
        to_send = len(response)

        # Send response in chunks.
        while to_send > 0:
            self.rawmsg('AUTHENTICATE', response[:RESPONSE_LIMIT])
            response = response[RESPONSE_LIMIT:]
            to_send -= RESPONSE_LIMIT

        # If our message fit exactly in SASL_RESPOSE_LIMIT-byte chunks, send an empty message to indicate we're done.
        if to_send == 0:
            self.rawmsg('AUTHENTICATE', EMPTY_MESSAGE)


    ## Capability callbacks.

    def on_capability_sasl_available(self):
        """ Check whether or not SASL is available. """
        if self.sasl_username and self.sasl_password:
            if puresasl:
                return True
            self.logger.warn('SASL credentials set but puresasl module not found: not initiating SASL authentication.')
        return False

    def on_capability_sasl_enabled(self):
        """ Start SASL authentication. """
        # Set a timeout handler.
        self._sasl_timer = threading.Timer(self.SASL_TIMEOUT, self._sasl_abort)
        self._sasl_timer.start()

        # Initialize SASL.
        self._sasl_start()

        # Tell caller we need more time, and to not end capability negotiation just yet.
        return cap.NEGOTIATING


    ## Message handlers.

    def on_raw_authenticate(self, source, params):
        """ Received part of the authentication challenge. """
        # Cancel timeout timer.
        self._sasl_timer.cancel()

        # Add response data.
        response = ' '.join(params)
        if response != EMPTY_MESSAGE:
            self._sasl_challenge += base64.b64decode(response)

        # If the response ain't exactly SASL_RESPONSE_LIMIT bytes long, it's the end. Process.
        if len(response) % RESPONSE_LIMIT > 0:
            self._sasl_respond()
        else:
            # Response not done yet. Restart timer.
            self._sasl_timer = threading.Timer(self.SASL_TIMEOUT, self._sasl_abort)
            self._sasl_timer.start()


    on_raw_900 = cap.CapabilityNegotiationSupport._ignored # You are now logged in as...

    def on_raw_903(self, source, params):
        """ SASL authentication successful. """
        self._sasl_end()

    def on_raw_904(self, source, params):
        """ Invalid mechanism or authentication failed. Abort SASL. """
        self._sasl_abort()

    def on_raw_905(self, source_params):
        """ Authentication failed. Abort SASL. """
        self._sasl_abort()

    on_raw_906 = cap.CapabilityNegotiationSupport._ignored # Completed registration while authenticating/registration aborted.
    on_raw_907 = cap.CapabilityNegotiationSupport._ignored # Already authenticated over SASL.



## run.py
# Run client.
import pydle
from . import _args

def main():
    client = _args.client_from_args('pydle', description='pydle IRC library.')
    client.handle_forever()

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages
import pydle

setup(
    name=pydle.__name__,
    version=pydle.__version__,
    packages=[
        'pydle',
        'pydle.features',
        'pydle.utils'
    ],
    extras_require={
        'SASL': 'pure-sasl >=0.1.6'   # for pydle.features.sasl
    },
    entry_points={
        'console_scripts': [
            'irccat = pydle.utils.irccat:main',
            'pydle = pydle.utils.run:main',
            'ipydle = pydle.utils.console:main'
        ]
    },

    author='Shiz',
    author_email='hi@shiz.me',
    url='https://github.com/Shizmob/pydle',
    keywords='irc library python3 compact flexible',
    description='A compact, flexible and standards-abiding IRC library for Python 3.',
    license=pydle.__license__,

    zip_safe=True,
    test_suite='tests'
)



## tls.py
# TLS support.
import ssl

from .. import client
from .. import connection
from .. import protocol

__all__ = [ 'TLSSupport' ]


class TLSSupport(client.BasicClient):
    """ TLS and STARTTLS support. """

    ## Internal overrides.

    def __init__(self, *args, tls_client_cert=None, tls_client_cert_key=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.tls_client_cert = tls_client_cert
        self.tls_client_cert_key = tls_client_cert_key

    def connect(self, hostname=None, port=None, password=None, encoding='utf-8', channels=[], tls=False, tls_verify=False):
        """ Connect to IRC server, optionally over TLS. """
        # Disconnect from current connection.
        if self.connected:
            self.disconnect()

        self.password = password
        self._autojoin_channels = channels

        if not port:
            if tls:
                port = protocol.DEFAULT_TLS_PORT
            else:
                port = protocol.DEFAULT_PORT

        # Create connection.
        self.connection = connection.Connection(hostname, port,
            tls=tls, tls_verify=tls_verify,
            tls_certificate_file=self.tls_client_cert,
            tls_certificate_keyfile=self.tls_client_cert_key,
            encoding=encoding)

        # Connect.
        self.connection.connect()

        # Set logger name.
        self.logger.name = self.__class__.__name__ + ':' + self.server_tag

        # Send STARTTLS if we're not on TLS already.
        if not tls:
            self.rawmsg('STARTTLS')

        # And initiate the IRC connection.
        self._register()


    ## Message callbacks.

    def on_raw_421(self, source, params):
        """ Hijack to ignore absence of STARTTLS support. """
        if params[0] == 'STARTTLS':
            return
        super().on_raw_421(source, params)

    def on_raw_451(self, source, params):
        """ Hijack to ignore absence of STARTTLS support. """
        if params[0] == 'STARTTLS':
            return
        super().on_raw_451(source, params)

    def on_raw_670(self, source, params):
        """ Got the OK from the server to start a TLS connection. Let's roll. """
        self.connection.tls = True
        self.connection.setup_tls()

    def on_raw_691(self, source, params):
        """ Error setting up TLS server-side. """
        self.logger.err('Server experienced error in setting up TLS, not proceeding with TLS setup: {}', params[0])
